/*

  $Id: gsm-api.c,v 1.52 2002/10/27 22:44:23 bozo Exp $

  G N O K I I

  A Linux/Unix toolset and driver for Nokia mobile phones.

  This file is part of gnokii.

  Gnokii is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Gnokii is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with gnokii; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Copyright (C) 1999, 2000 Hugh Blemings & Pavel Jank ml.

  Provides a generic API for accessing functions on the phone, wherever
  possible hiding the model specific details.

  The underlying code should run in it's own thread to allow communications to
  the phone to be run independantly of mailing code that calls these API
  functions.

  Unless otherwise noted, all functions herein block until they complete.  The
  functions themselves are defined in a structure in gsm-common.h.

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "misc.h"
#include "gsm-common.h"
#include "cfgreader.h"
#include "data/rlp-common.h"
#include "gsm-statemachine.h"
#include "phones/nk6510.h"
#include "phones/nk7110.h"
#include "phones/nk6100.h"
#include "phones/nk3110.h"
#include "phones/nk2110.h"


#if defined(WIN32) && defined(_USRDLL)

/*	GnokiiDll.cpp : Defines the entry point for the DLL application.
 *	
 *	We don't do anything special here (yet) but the code is needed
 *	in order to create a DLL.
 *
 */
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	/*	For now this is enough to satisfy the compiler
	 *	and linker. Currently no extra code is needed
	 *	for initializing or deinitializing of the DLL.
	 */

    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}

#endif	/* defined(WIN32) && defined(_USRDLL) */

GSM_Statemachine GSM_SM;
gn_error (*gn_gsm_f)(GSM_Operation op, GSM_Data *data, GSM_Statemachine *state);


/* Define pointer to the GSM_Information structure used by external code to
   obtain information that varies from model to model. This structure is also
   defined in gsm-common.h */
API GSM_Information *gn_gsm_info;

/* Initialise interface to the phone. Model number should be a string such as
   3810, 5110, 6110 etc. Device is the serial port to use e.g. /dev/ttyS0, the
   user must have write permission to the device. */
static gn_error register_phone(GSM_Phone *phone, char *model, char *setupmodel, GSM_Statemachine *sm)
{
	GSM_Data *data = NULL;
	GSM_Data *p_data;
	gn_error error = GN_ERR_UNKNOWNMODEL;

	if (setupmodel) {
		data = calloc(1, sizeof(GSM_Data));
		data->Model = setupmodel;
		p_data = data;
	} else {
		p_data = NULL;
	}
	if (strstr(phone->Info.Models, model) != NULL)
		error = phone->Functions(GOP_Init, p_data, sm);

	if (data) free(data);
	return error;
}

#define REGISTER_PHONE(x, y) { \
	extern GSM_Phone phone_##x; \
	if ((ret = register_phone(&phone_##x, model, y, sm)) != GN_ERR_UNKNOWNMODEL) \
		return ret; \
}

API gn_error gn_gsm_initialise(char *model, char *device, char *initlength,
				const char *connection, GSM_Statemachine *sm)
{
	gn_error ret;
	char *sms_timeout;

	if (!strcasecmp(connection, "serial"))
		sm->Link.ConnectionType = GCT_Serial;
	else if (!strcasecmp(connection, "dau9p"))
		sm->Link.ConnectionType = GCT_DAU9P;
	else if (!strcasecmp(connection, "dlr3p"))
		sm->Link.ConnectionType = GCT_DLR3P;
	else if (!strcasecmp(connection, "infrared"))
		sm->Link.ConnectionType = GCT_Infrared;
#ifdef HAVE_IRDA
	else if (!strcasecmp(connection, "irda"))
		sm->Link.ConnectionType = GCT_Irda;
#endif
#ifndef WIN32
	else if (!strcasecmp(connection, "tcp"))
		sm->Link.ConnectionType = GCT_TCP;
	else if (!strcasecmp(connection, "tekram"))
		sm->Link.ConnectionType = GCT_Tekram;
#endif
	else return GN_ERR_NOTSUPPORTED;

	sm->Link.InitLength = atoi(initlength);
	sms_timeout = gn_cfg_get(gn_cfg_info, "sms", "timeout");
	if (!sms_timeout) sm->Link.SMSTimeout = 100;
	else sm->Link.SMSTimeout = atoi(sms_timeout) * 10;
	memset(&sm->Link.PortDevice, 0, sizeof(sm->Link.PortDevice));
	strncpy(sm->Link.PortDevice, device, sizeof(sm->Link.PortDevice) - 1);

	REGISTER_PHONE(nokia_7110, NULL);
	REGISTER_PHONE(nokia_6510, NULL);
	REGISTER_PHONE(nokia_6100, NULL);
#if 0
	REGISTER_PHONE(nokia_3110, NULL);
#ifndef WIN32
	REGISTER_PHONE(nokia_2110, NULL);
	REGISTER_PHONE(dancall_2711, NULL);
#endif
#endif
	REGISTER_PHONE(fake, NULL);
	REGISTER_PHONE(at, model);
	REGISTER_PHONE(nokia_6160, NULL);

	return GN_ERR_UNKNOWNMODEL;
}
